//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Base/Vector/RotMatrix.h
//! @brief     Declares class RotMatrix.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_BASE_VECTOR_ROTMATRIX_H
#define BORNAGAIN_BASE_VECTOR_ROTMATRIX_H

#include <array>
#include <heinz/Vectors3D.h>
#include <optional>

//! Rotation matrix in three dimensions. Represents group SO(3).
//! Internal parameterization based on quaternions.

class RotMatrix {
public:
    //! Constructs unit transformation
    RotMatrix();

    //! Destructor
    ~RotMatrix() = default;

    //! Creates rotation around x-axis
    static RotMatrix AroundX(double phi);

    //! Creates rotation around y-axis
    static RotMatrix AroundY(double phi);

    //! Creates rotation around z-axis
    static RotMatrix AroundZ(double phi);

    //! Creates rotation defined by Euler angles
    static RotMatrix EulerZXZ(double alpha, double beta, double gamma);

    //! Calculates the Euler angles corresponding to the rotation
    std::array<double, 3> zxzEulerAngles() const;

    //! Returns the inverse transformation.
    RotMatrix getInverse() const;

    //! Return transformed vector _v_.
    template <class T>
    T transformed(const T& v) const;

    //! Return transformed vector _v_.
    template <class T>
    T transformedInverse(const T& v) const;

    //! Composes two transformations
    RotMatrix operator*(const RotMatrix&) const;

    //! Provides equality operator
    bool operator==(const RotMatrix&) const;

    //! Determine if the transformation is trivial (identity)
    bool isIdentity() const;

    bool isXRotation() const;
    bool isYRotation() const;
    bool isZRotation() const;

    std::optional<double> angleAroundCoordAxis(int iAxis) const;

private:
    RotMatrix(double x_, double y_, double z_, double s_);

    double x, y, z, s;
};

#endif // BORNAGAIN_BASE_VECTOR_ROTMATRIX_H
