//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Data/PowerfieldIterator.h
//! @brief     Defines and implements template class PowerfieldIterator.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif

#ifndef USER_API
#ifndef BORNAGAIN_DEVICE_DATA_POWERFIELDITERATOR_H
#define BORNAGAIN_DEVICE_DATA_POWERFIELDITERATOR_H

#include <cstddef>
#include <iterator>
#include <utility>

//! Iterator for underlying Powerfield container.

template <class TValue, class TContainer>
class PowerfieldIterator {
public:
    //! Empty constructor to comply with stl forward iterators
    PowerfieldIterator();

    //! constructor
    PowerfieldIterator(TContainer* p_output_data, size_t start_at_index = 0);

    //! templated copy construction
    template <class TValue2, class TContainer2>
    PowerfieldIterator(const PowerfieldIterator<TValue2, TContainer2>& other);

    //! non-templated copy construction
    PowerfieldIterator(const PowerfieldIterator<TValue, TContainer>& other);

    //! templated copy assignment
    template <class TValue2, class TContainer2>
    PowerfieldIterator<TValue, TContainer>&
    operator=(const PowerfieldIterator<TValue2, TContainer2>& right);

    //! non-templated copy asssignment
    PowerfieldIterator<TValue, TContainer>&
    operator=(const PowerfieldIterator<TValue, TContainer>& right);

    ~PowerfieldIterator();

    //! prefix increment
    PowerfieldIterator<TValue, TContainer>& operator++();

    //! postfix increment
    PowerfieldIterator<TValue, TContainer> operator++(int);

    //! retrieve current element
    TValue& operator*() const;

    //! pointer access
    TValue* operator->() const;

    //! Returns current index
    size_t getIndex() const { return m_current_index; }

    //! Returns container pointer
    TContainer* getContainer() const { return m_output_data; }

    //! Swaps iterators
    void swap(PowerfieldIterator<TValue, TContainer>& other);

    // typedefs for std::iterator_traits
    using iterator_category = std::forward_iterator_tag;
    using value_type = TValue;
    using difference_type = ptrdiff_t;
    using pointer_type = TValue*;
    using reference_type = TValue&;
    //#ifdef _MSC_VER
    using pointer = TValue*;
    using reference = TValue&;
    //#endif

private:
    size_t m_current_index;
    TContainer* m_output_data;
};

//! make Swappable
template <class TValue, class TContainer>
void swap(PowerfieldIterator<TValue, TContainer>& left,
          PowerfieldIterator<TValue, TContainer>& right)
{
    left.swap(right);
}

//! test for equality
template <class TValue1, class TContainer1, class TValue2, class TContainer2>
bool operator==(const PowerfieldIterator<TValue1, TContainer1>& left,
                const PowerfieldIterator<TValue2, TContainer2>& right);

//! test for inequality
template <class TValue1, class TContainer1, class TValue2, class TContainer2>
bool operator!=(const PowerfieldIterator<TValue1, TContainer1>& left,
                const PowerfieldIterator<TValue2, TContainer2>& right);

template <class TValue, class TContainer>
PowerfieldIterator<TValue, TContainer>::PowerfieldIterator()
    : m_current_index(0)
    , m_output_data(0)
{
}

template <class TValue, class TContainer>
PowerfieldIterator<TValue, TContainer>::PowerfieldIterator(TContainer* p_output_data,
                                                           size_t start_at_index)
    : m_current_index(start_at_index)
    , m_output_data(p_output_data)
{
}

template <class TValue, class TContainer>
template <class TValue2, class TContainer2>
PowerfieldIterator<TValue, TContainer>::PowerfieldIterator(
    const PowerfieldIterator<TValue2, TContainer2>& other)
    : m_current_index(0)
    , m_output_data(0)
{
    m_output_data = static_cast<TContainer*>(other.getContainer());
    m_current_index = other.getIndex();
}

template <class TValue, class TContainer>
PowerfieldIterator<TValue, TContainer>::PowerfieldIterator(
    const PowerfieldIterator<TValue, TContainer>& other)
    : m_current_index(0)
    , m_output_data(0)
{
    m_output_data = other.getContainer();
    m_current_index = other.getIndex();
}

template <class TValue, class TContainer>
template <class TValue2, class TContainer2>
PowerfieldIterator<TValue, TContainer>& PowerfieldIterator<TValue, TContainer>::operator=(
    const PowerfieldIterator<TValue2, TContainer2>& right)
{
    PowerfieldIterator<TValue, TContainer> copy(right);
    swap(copy);
    return *this;
}

template <class TValue, class TContainer>
PowerfieldIterator<TValue, TContainer>& PowerfieldIterator<TValue, TContainer>::operator=(
    const PowerfieldIterator<TValue, TContainer>& right)
{
    PowerfieldIterator<TValue, TContainer> copy(right);
    swap(copy);
    return *this;
}

template <class TValue, class TContainer>
PowerfieldIterator<TValue, TContainer>::~PowerfieldIterator() = default;

template <class TValue, class TContainer>
PowerfieldIterator<TValue, TContainer>& PowerfieldIterator<TValue, TContainer>::operator++()
{
    if (m_current_index < m_output_data->allocatedSize()) {
        ++m_current_index;
    }
    return *this;
}

template <class TValue, class TContainer>
PowerfieldIterator<TValue, TContainer> PowerfieldIterator<TValue, TContainer>::operator++(int /**/)
{
    PowerfieldIterator<TValue, TContainer> result(*this);
    this->operator++();
    return result;
}

template <class TValue, class TContainer>
TValue& PowerfieldIterator<TValue, TContainer>::operator*() const
{
    return (*m_output_data)[m_current_index];
}

template <class TValue, class TContainer>
TValue* PowerfieldIterator<TValue, TContainer>::operator->() const
{
    return &((*m_output_data)[m_current_index]);
}

template <class TValue, class TContainer>
void PowerfieldIterator<TValue, TContainer>::swap(PowerfieldIterator<TValue, TContainer>& other)
{
    std::swap(this->m_current_index, other.m_current_index);
    std::swap(this->m_output_data, other.m_output_data);
}

//! test for equality
template <class TValue1, class TContainer1, class TValue2, class TContainer2>
bool operator==(const PowerfieldIterator<TValue1, TContainer1>& left,
                const PowerfieldIterator<TValue2, TContainer2>& right)
{
    return left.getContainer() == right.getContainer() && left.getIndex() == right.getIndex();
}

//! test for inequality
template <class TValue1, class TContainer1, class TValue2, class TContainer2>
bool operator!=(const PowerfieldIterator<TValue1, TContainer1>& left,
                const PowerfieldIterator<TValue2, TContainer2>& right)
{
    return !(left == right);
}

#endif // BORNAGAIN_DEVICE_DATA_POWERFIELDITERATOR_H
#endif // USER_API
