//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Histo/HistoUtils.cpp
//! @brief     PyObvject forward declaration.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Histo/HistoUtils.h"
#include "Base/Axis/Bin.h"
#include "Base/Math/Numeric.h"
#include "Device/Data/DataUtils.h"
#include "Device/Histo/Histogram2D.h"
#include "Device/Histo/IOFactory.h"
#include "Device/Histo/SimulationResult.h"
#include <cmath>
#include <iostream>
#include <tspectrum.h> // third-party code, extracted from CERN ROOT (class TSpectrum2)

std::vector<std::pair<double, double>> DataUtils::Histo::FindPeaks(const Histogram2D& hist,
                                                                   double sigma,
                                                                   const std::string& option,
                                                                   double threshold)
{
    std::unique_ptr<Powerfield<double>> data(hist.createPowerfield());
    std::vector<std::vector<double>> arr = DataUtils::Array::createVector2D(*data);
    tspectrum::Spectrum2D spec;
    auto peaks = spec.find_peaks(arr, sigma, option, threshold);

    // coordinates of peaks in histogram axes units
    std::vector<std::pair<double, double>> result;

    for (const auto& p : peaks) {
        double row_value = p.first;
        double col_value = p.second;

        auto xaxis_index = static_cast<size_t>(col_value);
        size_t yaxis_index = hist.yAxis().size() - 1 - static_cast<size_t>(row_value);

        Bin1D xbin = hist.xAxis().bin(xaxis_index);
        Bin1D ybin = hist.yAxis().bin(yaxis_index);

        double dx = col_value - static_cast<size_t>(col_value);
        double dy = -1.0 * (row_value - static_cast<size_t>(row_value));

        double x = xbin.center() + xbin.binSize() * dx;
        double y = ybin.center() + ybin.binSize() * dy;

        result.emplace_back(x, y);
    }
    return result;
}

//! Returns sum of relative differences between each pair of elements:
//! (a, b) -> 2*abs(a - b)/(|a| + |b|)      ( and zero if  a=b=0 within epsilon )
double DataUtils::Histo::RelativeDifference(const SimulationResult& dat,
                                            const SimulationResult& ref)
{
    if (dat.size() != ref.size())
        throw std::runtime_error("Error in DataUtils::Histo::RelativeDifference: "
                                 "different number of elements");
    if (dat.empty())
        return 0.0;
    double sum_of_diff = 0.0;
    for (size_t i = 0; i < dat.size(); ++i)
        sum_of_diff += Numeric::GetRelativeDifference(dat[i], ref[i]);
    return sum_of_diff / dat.size();
}

double DataUtils::Histo::getRelativeDifference(const IHistogram& dat, const IHistogram& ref)
{
    return DataUtils::Data::relativeDataDifference(
        *std::unique_ptr<Powerfield<double>>(dat.getData().meanValues()),
        *std::unique_ptr<Powerfield<double>>(ref.getData().meanValues()));
}

//! Returns true is relative difference is below threshold; prints informative output
bool DataUtils::Histo::checkRelativeDifference(const Powerfield<double>& dat,
                                               const Powerfield<double>& ref,
                                               const double threshold)
{
    const std::unique_ptr<const IHistogram> histo(IHistogram::createHistogram(dat));
    if (histo->getMinimum() == 0 && histo->getMaximum() == 0) {
        std::cerr << "FAILED: simulated data set is empty" << std::endl;
        return false;
    }

    try {
        const double diff = DataUtils::Data::relativeDataDifference(dat, ref);
        if (diff > threshold) {
            std::cerr << "FAILED: relative deviation of dat from ref is " << diff
                      << ", above given threshold " << threshold << std::endl;
            return false;
        }
        if (diff)
            std::cerr << "- OK: relative deviation of dat from ref is " << diff
                      << ", within given threshold " << threshold << std::endl;
        else
            std::cout << "- OK: dat = ref\n";
        return true;
    } catch (...) {
        return false;
    }
}

bool DataUtils::Histo::agreesWithReference(const SimulationResult& dat,
                                           const std::string& refFileName, double tol)
{
    std::unique_ptr<Powerfield<double>> refDat{IOFactory::readPowerfield(refFileName)};
    if (!refDat) {
        std::cerr << "Could not read reference data from file " << refFileName << std::endl;
        return false;
    }
    std::unique_ptr<Powerfield<double>> datDat(dat.data());
    return DataUtils::Histo::checkRelativeDifference(*datDat, *refDat, tol);
}

bool DataUtils::Histo::filesAgree(const std::string& datFileName, const std::string& refFileName,
                                  double tol)
{
    std::unique_ptr<Powerfield<double>> datDat;
    try {
        datDat.reset(IOFactory::readPowerfield(datFileName));
    } catch (const std::runtime_error& ex) {
        std::cerr << "File comparison: Could not read data from file " << datFileName
                  << ". Runtime error: " << ex.what() << std::endl;
    } catch (const std::exception& ex) {
        std::cerr << "File comparison: Could not read data from file " << datFileName
                  << ". Exception: " << ex.what() << std::endl;
    } catch (...) {
        std::cerr << "File comparison: Could not read data from file " << datFileName
                  << ". Unknown exception." << std::endl;
        return false;
    }
    ASSERT(datDat);

    std::unique_ptr<Powerfield<double>> refDat;
    try {
        refDat.reset(IOFactory::readPowerfield(refFileName));
    } catch (...) {
        std::cerr << "File comparison: Could not read reference data from file " << refFileName
                  << std::endl;
        return false;
    }
    ASSERT(refDat);

    return DataUtils::Histo::checkRelativeDifference(*datDat, *refDat, tol);
}
