//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Job/MinimizerItem.h
//! @brief     Defines MinimizerItem class
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_JOB_MINIMIZERITEM_H
#define BORNAGAIN_GUI_MODEL_JOB_MINIMIZERITEM_H

#include "GUI/Model/BaseItem/SessionItem.h"
#include "GUI/Model/Descriptor/DoubleDescriptor.h"
#include "GUI/Model/Descriptor/SelectionDescriptor.h"
#include "GUI/Model/Descriptor/UIntDescriptor.h"
#include <variant>

class IMinimizer;
class ObjectiveMetric;

using ValueDescriptors =
    QList<std::variant<DoubleDescriptor, UIntDescriptor, SelectionDescriptor<QString>>>;

//! The MinimizerItem class is the base item to hold minimizer settings.

class BA_CORE_API_ MinimizerItem : public SessionItem {
public:
    virtual std::unique_ptr<IMinimizer> createMinimizer() const = 0;

    virtual ValueDescriptors valueDescriptorsForUI() const { return {}; }

protected:
    explicit MinimizerItem(const QString& model_type);
};

//! The MinimizerContainerItem class holds collection of minimizers.

class MinimizerContainerItem : public MinimizerItem {
private:
    static constexpr auto P_MINIMIZERS{"Minimizer"};
    static constexpr auto P_METRIC{"Objective metric"};
    static constexpr auto P_NORM{"Norm function"};

public:
    static constexpr auto M_TYPE{"MinimizerContainer"};

    MinimizerContainerItem();

    SelectionDescriptor<MinimizerItem*> minimizers() const;
    SelectionDescriptor<QString> normFunction() const;
    SelectionDescriptor<QString> objectiveMetric() const;

    std::unique_ptr<IMinimizer> createMinimizer() const override;
    std::unique_ptr<ObjectiveMetric> createMetric() const;
};

//! The MinuitMinimizerItem class represents settings for ROOT Minuit2 minimizer.

class MinuitMinimizerItem : public MinimizerItem {
private:
    static constexpr auto P_ALGORITHMS{"Algorithms"};
    static constexpr auto P_STRATEGY{"Strategy"};
    static constexpr auto P_ERRORDEF{"ErrorDef"};
    static constexpr auto P_TOLERANCE{"Tolerance"};
    static constexpr auto P_PRECISION{"Precision"};
    static constexpr auto P_MAXFUNCTIONCALLS{"MaxFunctionCalls"};

public:
    static constexpr auto M_TYPE{"Minuit2"};

    MinuitMinimizerItem();
    std::unique_ptr<IMinimizer> createMinimizer() const override;
    ValueDescriptors valueDescriptorsForUI() const override;
};

//! The GSLMinimizerItem class represents settings for GSL MultiMin minimizer family.

class GSLMultiMinimizerItem : public MinimizerItem {
private:
    static constexpr auto P_ALGORITHMS{"Algorithms"};
    static constexpr auto P_MAXITERATIONS{"MaxIterations"};

public:
    static constexpr auto M_TYPE{"GSLMultiMin"};

    GSLMultiMinimizerItem();
    std::unique_ptr<IMinimizer> createMinimizer() const override;
    ValueDescriptors valueDescriptorsForUI() const override;
};

//! The GeneticMinimizerItem class represents settings for TMVA/Genetic minimizer.

class GeneticMinimizerItem : public MinimizerItem {
private:
    static constexpr auto P_TOLERANCE{"Tolerance"};
    static constexpr auto P_MAXITERATIONS{"MaxIterations"};
    static constexpr auto P_POPULATIONSIZE{"PopSize"};
    static constexpr auto P_RANDOMSEED{"RandomSeed"};

public:
    static constexpr auto M_TYPE{"Genetic"};

    GeneticMinimizerItem();
    std::unique_ptr<IMinimizer> createMinimizer() const override;
    ValueDescriptors valueDescriptorsForUI() const override;
};

//! The SimAnMinimizerItem class represents settings for GSL's simulated annealing minimizer.

class SimAnMinimizerItem : public MinimizerItem {
private:
    static constexpr auto P_MAXITERATIONS{"MaxIterations"};
    static constexpr auto P_ITERATIONSTEMP{"IterationsAtTemp"};
    static constexpr auto P_STEPSIZE{"StepSize"};
    static constexpr auto P_BOLTZMANN_K{"k"};
    static constexpr auto P_BOLTZMANN_TINIT{"t_init"};
    static constexpr auto P_BOLTZMANN_MU{"mu"};
    static constexpr auto P_BOLTZMANN_TMIN{"t_min"};

public:
    static constexpr auto M_TYPE{"GSLSimAn"};

    SimAnMinimizerItem();

    std::unique_ptr<IMinimizer> createMinimizer() const override;
    ValueDescriptors valueDescriptorsForUI() const override;
};

//! The GSLLMAMinimizerItem class represents settings for GSL's version of Levenberg-Marquardt.

class GSLLMAMinimizerItem : public MinimizerItem {
private:
    static constexpr auto P_TOLERANCE{"Tolerance"};
    static constexpr auto P_MAXITERATIONS{"MaxIterations"};

public:
    static constexpr auto M_TYPE{"GSLLMA"};

    GSLLMAMinimizerItem();
    std::unique_ptr<IMinimizer> createMinimizer() const override;
    ValueDescriptors valueDescriptorsForUI() const override;
};

#endif // BORNAGAIN_GUI_MODEL_JOB_MINIMIZERITEM_H
