//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Tool/WidgetUtils.h
//! @brief     Defines GUI::Util namespace
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2022
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_VIEW_TOOL_WIDGETUTILS_H
#define BORNAGAIN_GUI_VIEW_TOOL_WIDGETUTILS_H

#include "GUI/Model/Descriptor/SelectionDescriptor.h"
#include "GUI/Support/Type/Unit.h"
#include "GUI/View/Common/CustomEventFilters.h"
#include <QComboBox>
#include <variant>

class QSpinBox;
class UIntDescriptor;
class DoubleSpinBox;
class DoubleDescriptor;
class QFormLayout;
class ScientificSpinBox;

namespace GUI::Util {

//! Create a combo box with the information found in a selection descriptor.
//!
//! The combo will be filled with the available options and will get the found tooltip.
//! The current index will be set according to the current index in the selection.
//! Furthermore, the combo box will prohibit accidental changes by the mouse wheel. Otherwise it
//! would be dangerous if the combo is on a scrollable form - unintended and unnoticed changes would
//! take place when just scrolling through the form.
//!
//! Changes in the combobox will be notified to the SelectionDescriptor already. The additional (and
//! optional) slot can be used to be notified about an already executed change.
//!
template <typename T>
QComboBox* createSelectionCombo(QWidget* parent, const SelectionDescriptor<T> d,
                                std::function<void(int)> slot = nullptr)
{
    auto* combo = new QComboBox(parent);
    combo->addItems(d.options);
    combo->setMaxCount(d.options.size());
    combo->setToolTip(d.tooltip);
    combo->setCurrentIndex(d.currentIndex());
    WheelEventEater::install(combo);

    QObject::connect(combo, qOverload<int>(&QComboBox::currentIndexChanged),
                     [=](int newIndex) { d.setCurrentIndex(newIndex); });

    if (slot)
        QObject::connect(combo, qOverload<int>(&QComboBox::currentIndexChanged),
                         [=](int newIndex) { slot(newIndex); });

    return combo;
}

//! Create a spin box with the information found in a UIntDescriptor.
//!
//! The spin box will be fully initialized (tooltip, limits, current value, size policy).
//! Furthermore, the spin box will prohibit accidental changes by the mouse wheel. Otherwise it
//! would be dangerous if the spin box is on a scrollable form - unintended and unnoticed changes
//! would take place when just scrolling through the form.
//!
//! No connections to update the descriptor will be established! Therefore changes in the spin box
//! will *not* be notified to the descriptor. The additional (and optional) slot can be used to be
//! notified about a value change.
QSpinBox* createSpinBox(QWidget* parent, const UIntDescriptor& d,
                        std::function<void(uint)> slot = nullptr);

//! Create a label and a spin box with the information found in a UIntDescriptor and place them in a
//! row in a form layout.
//!
//! The label will also contain the unit, if present.
//! Regarding the spin box creation see the method above.
QSpinBox* createSpinBox(QFormLayout* parentLayout, const UIntDescriptor& d);

//! Create a label and a spin box with the information found in a DoubleDescriptor and place them in
//! a row in a form layout.
//!
//! The spin box will be fully initialized (tooltip, limits, current value, unit, size policy).
//! Furthermore, the spin box will prohibit accidental changes by the mouse wheel. Otherwise it
//! would be dangerous if the spin box is on a scrollable form - unintended and unnoticed changes
//! would take place when just scrolling through the form.
//!
//! No connections to update the descriptor will be established! Therefore changes in the spin box
//! will *not* be notified to the descriptor. The additional (and optional) slot can be used to be
//! notified about a value change (the notified new value is in the base unit of the
//! DoubleDescriptor).
DoubleSpinBox* createSpinBox(QFormLayout* parentLayout, const DoubleDescriptor& d,
                             std::function<void(double)> slot = nullptr);

//! Create a label and a scientific spin box with the information found in a DoubleDescriptor and
//! place them in a row in a form layout.
//!
//! The spin box will be fully initialized (tooltip, limits, current value, unit, size policy).
//! Furthermore, the spin box will prohibit accidental changes by the mouse wheel. Otherwise it
//! would be dangerous if the spin box is on a scrollable form - unintended and unnoticed changes
//! would take place when just scrolling through the form.
//!
//! No connections to update the descriptor will be established! Therefore changes in the spin box
//! will *not* be notified to the descriptor. The additional (and optional) slot can be used to be
//! notified about a value change.
ScientificSpinBox* createScientificSpinBox(QFormLayout* parentLayout, const DoubleDescriptor& d,
                                           std::function<void(double)> slot = nullptr);

//! Create a label with an optional unit in brackets.
//!
//! No trailing ':' will be appended
QString labelWithUnit(const QString& label, std::variant<QString, Unit> unit);

//! Create a label with an optional unit in brackets, both taken from the given descriptor.
//!
//! No trailing ':' will be appended
QString labelWithUnit(const DoubleDescriptor& d);

} // namespace GUI::Util

#endif // BORNAGAIN_GUI_VIEW_TOOL_WIDGETUTILS_H
