#include "Base/Vector/RotMatrix.h"
#include "Tests/GTestWrapper/google_test.h"

class RotMatrixTest : public ::testing::Test {
protected:
    const double epsilon = 1e-12;

    const double w0 = M_PI / 5;
    const double w1 = M_PI / 7;
    const double w2 = M_PI / 11;

    const RotMatrix mEul = RotMatrix::EulerZXZ(w0, w1, w2);

    void InversionTest(const RotMatrix& mRot, const R3& a0)
    {

        const RotMatrix mInv = mRot.getInverse();

        const R3 a1 = mRot.transformed(a0);
        const R3 a2 = mInv.transformed(a1);

        EXPECT_NEAR(a2.x(), a0.x(), epsilon);
        EXPECT_NEAR(a2.y(), a0.y(), epsilon);
        EXPECT_NEAR(a2.z(), a0.z(), epsilon);
    }
};


TEST_F(RotMatrixTest, RotateY)
{
    const R3 a(std::sqrt(3.) / 2., 2., 0.5);
    const RotMatrix m2 = RotMatrix::AroundY(M_PI / 6.);
    const R3 v = m2.transformed(a);

    EXPECT_NEAR(v.x(), 1.0, epsilon);
    EXPECT_DOUBLE_EQ(v.y(), 2.0);
    EXPECT_NEAR(v.z(), 0.0, epsilon);
}

TEST_F(RotMatrixTest, RotateZ)
{
    const R3 a(0.5, std::sqrt(3.) / 2., 2.);
    const RotMatrix m3 = RotMatrix::AroundZ(M_PI / 6.);
    const R3 v = m3.transformed(a);

    EXPECT_NEAR(v.x(), 0.0, epsilon);
    EXPECT_NEAR(v.y(), 1.0, epsilon);
    EXPECT_DOUBLE_EQ(v.z(), 2.0);

    const RotMatrix m4 = m3.getInverse();
    const R3 w = m4.transformed(v);

    EXPECT_NEAR(w.x(), a.x(), epsilon);
    EXPECT_NEAR(w.y(), a.y(), epsilon);
    EXPECT_NEAR(w.z(), a.z(), epsilon);
}

TEST_F(RotMatrixTest, RecoverEulerAngles)
{
    auto angles = mEul.zxzEulerAngles();

    EXPECT_NEAR(angles[0], w0, epsilon);
    EXPECT_NEAR(angles[1], w1, epsilon);
    EXPECT_NEAR(angles[2], w2, epsilon);
}

TEST_F(RotMatrixTest, InvertXMatrix)
{
    InversionTest(RotMatrix::AroundX(M_PI / 7.), R3(4, 5, 6));
}

TEST_F(RotMatrixTest, InvertYMatrix)
{
    InversionTest(RotMatrix::AroundY(M_PI / 7.), R3(4, 5, 6));
}

TEST_F(RotMatrixTest, InvertZMatrix)
{
    InversionTest(RotMatrix::AroundZ(M_PI / 7.), R3(4, 5, 6));
}

TEST_F(RotMatrixTest, InvertEulerMatrix)
{
    InversionTest(mEul, R3(3, 4, 7));
}
