#include "Base/Axis/Bin.h"
#include "Base/Axis/VariableBinAxis.h"
#include "Device/Data/DataUtils.h"
#include "Tests/GTestWrapper/google_test.h"
#include <algorithm>

class PowerfieldTest : public ::testing::Test {
protected:
    PowerfieldTest();

    Powerfield<double> db_data;
};

PowerfieldTest::PowerfieldTest()
    : db_data(FixedBinAxis("angle", 20, 0.0, 0.1), FixedBinAxis("length", 10, 0.0, 0.5))
{
    Powerfield<double>::iterator it = db_data.begin();
    for (size_t i = 0; i < 200; ++i) {
        *it = (double)i;
        ++it;
    }
}

TEST_F(PowerfieldTest, SizeAfterAddingAxes)
{
    EXPECT_EQ(200u, db_data.allocatedSize());
}

TEST_F(PowerfieldTest, DataInitialization)
{
    std::vector<unsigned> coordinates;
    coordinates.push_back(11);
    coordinates.push_back(4);
    EXPECT_DOUBLE_EQ(114., db_data[db_data.toGlobalIndex(coordinates)]);
}

TEST_F(PowerfieldTest, DataCopying)
{
    Powerfield<double> data1(FixedBinAxis("axis0", 10, 0., 10.),
                             FixedBinAxis("axis1", 10, 0., 10.));
    Powerfield<double> data2(FixedBinAxis("axis0", 10, 0., 10.),
                             FixedBinAxis("axis1", 10, 1., 10.));

    EXPECT_TRUE(data1.hasSameSizes(data2));
    EXPECT_FALSE(data1.hasSameShape(data2));

    data2.copyFrom(data1);
    EXPECT_TRUE(data1.hasSameSizes(data2));
    EXPECT_TRUE(data1.hasSameShape(data2));

    data1.setAllTo(10);
    data2.copyFrom(data1);
    EXPECT_TRUE(data1.totalSum() == data2.totalSum());
    EXPECT_TRUE(data1.hasSameSizes(data2));
    EXPECT_TRUE(data1.hasSameShape(data2));
    EXPECT_TRUE(data2.hasSameSizes(data1));
    EXPECT_TRUE(data2.hasSameShape(data1));
    EXPECT_TRUE(data1.hasSameSizes(data1));
    EXPECT_TRUE(data1.hasSameShape(data1));
    EXPECT_TRUE(data2.hasSameSizes(data2));
    EXPECT_TRUE(data2.hasSameShape(data2));
}

// y |
// --------------------------------------------
// 1 | 1   3   5   7   9   11  13  15  17  19 |
// 0 | 0   2   4   6   8   10  12  14  16  18 |
// --------------------------------------------
//   | 0   1   2   3   4   5   6   7   8   9  | x
TEST_F(PowerfieldTest, ValueOfAxis)
{
    Powerfield<double> data{FixedBinAxis("axis1", 10, 0., 10.), FixedBinAxis("axis2", 2, 0., 10.)};
    EXPECT_EQ(0.5, data.getAxisValue(0, "axis1"));
    EXPECT_EQ(0.5, data.getAxisValue(1, "axis1"));
    EXPECT_EQ(1.5, data.getAxisValue(2, "axis1"));
    EXPECT_EQ(1.5, data.getAxisValue(3, "axis1"));
    EXPECT_EQ(9.5, data.getAxisValue(18, "axis1"));
    EXPECT_EQ(9.5, data.getAxisValue(19, "axis1"));

    EXPECT_EQ(2.5, data.getAxisValue(0, "axis2"));
    EXPECT_EQ(7.5, data.getAxisValue(1, "axis2"));
    EXPECT_EQ(2.5, data.getAxisValue(2, "axis2"));
    EXPECT_EQ(7.5, data.getAxisValue(3, "axis2"));
    EXPECT_EQ(2.5, data.getAxisValue(18, "axis2"));
    EXPECT_EQ(7.5, data.getAxisValue(19, "axis2"));

    std::vector<double> coordinates = data.getAxesValues(18);
    EXPECT_EQ(9.5, coordinates[0]);
    EXPECT_EQ(2.5, coordinates[1]);
}

// y |
// --------------------------------------------
// 1 | 1   3   5   7   9   11  13  15  17  19 |
// 0 | 0   2   4   6   8   10  12  14  16  18 |
// --------------------------------------------
//   | 0   1   2   3   4   5   6   7   8   9  | x
TEST_F(PowerfieldTest, GetAxisBin)
{
    Powerfield<double> data{FixedBinAxis("axis1", 10, 0., 10.), FixedBinAxis("axis2", 2, 0., 10.)};

    EXPECT_EQ(0.5, data.getAxisBin(0, "axis1").center());
    EXPECT_EQ(0.5, data.getAxisBin(1, "axis1").center());
    EXPECT_EQ(9.5, data.getAxisBin(18, "axis1").center());
    EXPECT_EQ(9.5, data.getAxisBin(19, "axis1").center());
    EXPECT_EQ(2.5, data.getAxisBin(0, "axis2").center());
    EXPECT_EQ(7.5, data.getAxisBin(1, "axis2").center());
    EXPECT_EQ(2.5, data.getAxisBin(18, "axis2").center());
    EXPECT_EQ(7.5, data.getAxisBin(19, "axis2").center());
}

TEST_F(PowerfieldTest, SetCleared)
{
    db_data.clear();
    db_data.setAllTo(1.0);
    EXPECT_EQ(db_data[0], 1.0);
}
