#include "Device/Histo/IOFactory.h"
#include "GUI/Model/Data/IntensityDataItem.h"
#include "GUI/Model/Data/JobItemUtils.h"
#include "GUI/Model/Data/RealDataItem.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Model/ApplicationModels.h"
#include "GUI/Model/Model/JobFunctions.h"
#include "GUI/Model/Model/JobModel.h"
#include "GUI/Model/Model/RealDataModel.h"
#include "GUI/Model/Project/IOService.h"
#include "GUI/Support/IO/ImportDataInfo.h"
#include "GUI/Support/IO/ProjectUtils.h"
#include "GUI/Util/Error.h"
#include "GUI/Util/Path.h"
#include "Tests/GTestWrapper/google_test.h"
#include "Tests/Unit/GUI/Utils.h"
#include <QTest>
#include <memory>

class TestIOService : public ::testing::Test {
protected:
    Powerfield<double> m_data{FixedBinAxis("x", 10, 0.0, 1.0), FixedBinAxis("y", 10, -1.0, 1.0)};
};


//! Test methods for retrieving nonXML data.

TEST_F(TestIOService, nonXMLData)
{
    ApplicationModels models;

    // initial state
    auto dataItems = models.nonXMLItems();
    EXPECT_EQ(dataItems.size(), 0);

    // adding RealDataItem
    auto* realData = models.realDataModel()->insertIntensityDataItem();
    realData->setPowerfield(UTest::GUI::createData().release());
    EXPECT_EQ(models.realDataModel()->nonXMLItems().size(), 1);

    // adding JobItem
    auto* jobItem = models.jobModel()->insertItem<JobItem>();
    auto* dataItem = jobItem->setDataType<IntensityDataItem>();
    EXPECT_EQ(models.jobModel()->nonXMLItems().size(), 1);

    // adding RealDataItem to jobItem
    RealDataItem* realData2 = jobItem->createRealDataItem();
    EXPECT_EQ(models.jobModel()->nonXMLItems().size(), 1);
    realData2->setPowerfield(UTest::GUI::createData(0.0, UTest::GUI::DIM::D1).release());
    EXPECT_EQ(models.jobModel()->nonXMLItems().size(), 2);

    // checking data items of PowerfieldIOService
    PowerfieldIOService service(&models);
    EXPECT_EQ(service.dataInterfaces().size(), 3);

    // checking data items of ApplicationModels
    dataItems = models.nonXMLItems();
    EXPECT_EQ(dataItems.size(), 3);
    EXPECT_EQ(dataItems.indexOf(realData->dataItem()), 0);
    EXPECT_EQ(dataItems.indexOf(dataItem), 1);
    EXPECT_EQ(dataItems.indexOf(realData2->dataItem()), 2);

    // Replacing the data inside RealDataItem with the data of the same dimensions
    realData->setPowerfield(UTest::GUI::createData(2.0).release());
    EXPECT_EQ(models.realDataModel()->nonXMLItems().size(), 1);

    // Replacing the data inside RealDataItem with the data of different dimensions
    auto data = UTest::GUI::createData(3.0, UTest::GUI::DIM::D1);
    EXPECT_THROW(dynamic_cast<RealDataItem*>(realData)->setPowerfield(data.get()), Error);
    EXPECT_EQ(models.realDataModel()->nonXMLItems().size(), 1);
}

//! Tests PowerfieldSaveInfo class intended for storing info about the last save.

TEST_F(TestIOService, SaveInfo)
{
    SessionModel model("TempModel");
    auto* item = model.insertItem<IntensityDataItem>();

    item->setLastModified(QDateTime::currentDateTime());

    const int nap_time(10);
    QTest::qSleep(nap_time);

    PowerfieldSaveInfo info = PowerfieldSaveInfo::createSaved(item);
    EXPECT_FALSE(info.wasModifiedSinceLastSave());

    QTest::qSleep(nap_time);
    item->setLastModified(QDateTime::currentDateTime());
    EXPECT_TRUE(info.wasModifiedSinceLastSave());
}

//! Tests PowerfieldDirHistory class intended for storing save history of several
//! IntensityDataItems in a directory.

TEST_F(TestIOService, DirHistory)
{
    SessionModel model("TempModel");
    auto* item1 = model.insertItem<IntensityDataItem>();
    auto* item2 = model.insertItem<IntensityDataItem>();
    item1->setPowerfield(m_data.clone());
    item1->setLastModified(QDateTime::currentDateTime());
    item2->setLastModified(QDateTime::currentDateTime());

    // empty history
    PowerfieldDirHistory history;
    EXPECT_FALSE(history.contains(item1));
    // non-existing item is treated as modified
    EXPECT_TRUE(history.wasModifiedSinceLastSave(item1));

    // Saving item in a history
    history.markAsSaved(item1);
    history.markAsSaved(item2);

    EXPECT_TRUE(history.contains(item1));
    // Empty DataItems are not added to history:
    EXPECT_FALSE(history.contains(item2));
    EXPECT_FALSE(history.wasModifiedSinceLastSave(item1));

    // Attempt to save same item second time
    EXPECT_THROW(history.markAsSaved(item1), Error);

    // Modifying item
    QTest::qSleep(10);
    item1->setLastModified(QDateTime::currentDateTime());

    EXPECT_TRUE(history.wasModifiedSinceLastSave(item1));
}

//! Tests PowerfieldIOHistory class (save info for several independent directories).

TEST_F(TestIOService, IOHistory)
{
    SessionModel model("TempModel");
    auto* item1 = model.insertItem<IntensityDataItem>();

    auto* item2 = model.insertItem<IntensityDataItem>();
    item1->setPowerfield(m_data.clone());
    item2->setPowerfield(m_data.clone());

    item1->setLastModified(QDateTime::currentDateTime());
    item2->setLastModified(QDateTime::currentDateTime());

    PowerfieldDirHistory dirHistory1;
    dirHistory1.markAsSaved(item1);
    dirHistory1.markAsSaved(item2);

    PowerfieldDirHistory dirHistory2;
    dirHistory2.markAsSaved(item1);

    // Modifying item
    QTest::qSleep(10);
    item1->setLastModified(QDateTime::currentDateTime());

    // Creating two histories
    PowerfieldIOHistory history;
    history.setHistory("dir1", dirHistory1);
    history.setHistory("dir2", dirHistory2);

    EXPECT_TRUE(history.wasModifiedSinceLastSave("dir1", item1));
    EXPECT_TRUE(history.wasModifiedSinceLastSave("dir2", item1));

    EXPECT_FALSE(history.wasModifiedSinceLastSave("dir1", item2));
    EXPECT_TRUE(history.wasModifiedSinceLastSave("dir2", item2)
                == true); // since item2 doesn't exist

    // asking info for some non-existing directory
    EXPECT_THROW(history.wasModifiedSinceLastSave("dir3", item1), Error);
}

//! Testing saving abilities of PowerfieldIOService class.

TEST_F(TestIOService, IOService)
{
    const QString projectDir("test_PowerfieldIOService");
    UTest::GUI::create_dir(projectDir);

    const double value1(1.0), value2(2.0), value3(3.0);

    ApplicationModels models;
    RealDataItem* realData1 = UTest::GUI::createRealData("data1", *models.realDataModel(), value1);
    RealDataItem* realData2 = UTest::GUI::createRealData("data2", *models.realDataModel(), value2);

    realData1->updateNonXMLDataFileNames();
    realData2->updateNonXMLDataFileNames();

    // Saving first time
    PowerfieldIOService service(&models);
    service.save(projectDir);
    QTest::qSleep(10);

    // Checking existence of data on disk
    QString fname1 = "./" + projectDir + "/realdata_data1_0.int.gz";
    QString fname2 = "./" + projectDir + "/realdata_data2_0.int.gz";
    EXPECT_TRUE(QFile::exists(fname1));
    EXPECT_TRUE(QFile::exists(fname2));

    // Reading data from disk, checking it is the same
    std::unique_ptr<Powerfield<double>> dataOnDisk1(
        IOFactory::readPowerfield(fname1.toStdString()));
    std::unique_ptr<Powerfield<double>> dataOnDisk2(
        IOFactory::readPowerfield(fname2.toStdString()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk1, *realData1->dataItem()->getPowerfield()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *realData2->dataItem()->getPowerfield()));

    // Modifying data and saving the project.
    realData2->setPowerfield(UTest::GUI::createData(value3).release());
    service.save(projectDir);
    QTest::qSleep(10);

    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk1, *realData1->dataItem()->getPowerfield()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *realData2->dataItem()->getPowerfield())
                == false);
    // checking that data on disk has changed
    dataOnDisk2.reset(IOFactory::readPowerfield(fname2.toStdString()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *realData2->dataItem()->getPowerfield()));

    // Renaming RealData and check that file on disk changed the name
    realData2->setDataName("data2new");
    realData2->updateNonXMLDataFileNames();
    service.save(projectDir);
    QTest::qSleep(10);

    QString fname2new = "./" + projectDir + "/realdata_data2new_0.int.gz";
    EXPECT_TRUE(QFile::exists(fname2new));
    EXPECT_TRUE(UTest::GUI::isTheSame(fname2new, *realData2->dataItem()->getPowerfield()));

    // Check that file with old name was removed.
    EXPECT_FALSE(QFile::exists(fname2));
}

TEST_F(TestIOService, RealDataItemWithNativeData)
{
    ApplicationModels models;

    // initial state
    auto dataItems = models.nonXMLItems();
    EXPECT_EQ(dataItems.size(), 0);

    // adding RealDataItem
    auto* realData = models.realDataModel()->insertIntensityDataItem();
    EXPECT_EQ(models.realDataModel()->nonXMLItems().size(), 1);

    ImportDataInfo import_data(std::unique_ptr<Powerfield<double>>(m_data.clone()), "nbins");
    realData->setImportData(std::move(import_data));
    // TODO: setImportData tested here with 2D data while main code uses it for 1D only.

    EXPECT_TRUE(realData->isIntensityData());
    EXPECT_EQ(models.realDataModel()->nonXMLItems().size(), 2);
    realData->setDataName("RealData");

    // adding JobItem
    auto* jobItem = models.jobModel()->insertItem<JobItem>();
    jobItem->setIdentifier(QUuid::createUuid().toString());
    jobItem->setDataType<IntensityDataItem>();
    EXPECT_EQ(models.jobModel()->nonXMLItems().size(), 1);

    // copying RealDataItem to JobItem
    GUI::Model::JobFunctions::copyRealDataItem(jobItem, realData);
    EXPECT_EQ(models.jobModel()->nonXMLItems().size(), 3);

    // checking data items of PowerfieldIOService
    PowerfieldIOService service(&models);
    EXPECT_EQ(service.dataInterfaces().size(), 5);

    const QString projectDir("test_NativeData");
    UTest::GUI::create_dir(projectDir);

    // Saving
    service.save(projectDir);
    QTest::qSleep(10);

    // Checking existance of data on disk
    auto* data1 = realData->dataItem();
    auto* data2 = realData->nativeData();
    auto* data3 = jobItem->realDataItem()->dataItem();
    auto* data4 = jobItem->realDataItem()->nativeData();

    QString fname1 = "./" + projectDir + "/" + data1->fileName();
    QString fname2 = "./" + projectDir + "/" + data2->fileName();
    QString fname3 = "./" + projectDir + "/" + data3->fileName();
    QString fname4 = "./" + projectDir + "/" + data4->fileName();

    EXPECT_TRUE(QFile::exists(fname1));
    EXPECT_TRUE(QFile::exists(fname2));
    EXPECT_TRUE(QFile::exists(fname3));
    EXPECT_TRUE(QFile::exists(fname4));

    auto readData = [](const QString& filename) {
        return std::unique_ptr<Powerfield<double>>(
            IOFactory::readPowerfield(filename.toStdString()));
    };

    // Reading data from disk, checking it is the same
    auto dataOnDisk1 = readData(fname1);
    auto dataOnDisk2 = readData(fname2);
    auto dataOnDisk3 = readData(fname3);
    auto dataOnDisk4 = readData(fname4);
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk1, *data1->getPowerfield()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *data2->getPowerfield()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk3, *data3->getPowerfield()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk4, *data4->getPowerfield()));
}
