#include "Base/Const/Units.h"
#include "Base/Math/Constants.h"
#include "Sample/HardParticle/HardParticles.h"
#include "Sample/Scattering/Rotations.h"
#include "Tests/GTestWrapper/google_test.h"

class FormFactorBasicTest : public ::testing::Test {
protected:
    void test_eps_q(const IFormFactor* p, C3 qdir, double eps) const
    {
        C3 q = eps * qdir;
        complex_t ff = p->formfactor_at_bottom(q);
        // std::cout<<"q="<<q<<" -> "<<std::setprecision(16)<<" ff0="<<V<<", ff ="<<ff<<"\n";
        EXPECT_LE(real(ff), V * (1 + 4e-16));
        if (R * R * R < V / 20 || R * R * R > 20 * V)
            // very excentric shape, the following tests cannot be expected to pass
            return;
        EXPECT_GT(real(ff), V * (1 - std::max(3e-16, 2 * eps * R * eps * R)));
        EXPECT_LT(std::abs(imag(ff)), 2 * eps * V * R);
    }
    void test_small_q(const IFormFactor* p, complex_t x, complex_t y, complex_t z) const
    {
        C3 q(x, y, z);
        test_eps_q(p, q, 1e-14);
        test_eps_q(p, q, 1e-11);
        test_eps_q(p, q, 1e-8);
        test_eps_q(p, q, 1e-5);
    }
    void test_ff(const IFormFactor* p)
    {
        complex_t ff0 = p->formfactor_at_bottom(C3(0., 0., 0.));
        EXPECT_EQ(imag(ff0), 0.);
        V = real(ff0);
        EXPECT_NEAR(p->volume(), V, 3e-16 * V);

        R = p->radialExtension();
        if (R * R * R < V / 20 || R * R * R > 20 * V) {
            std::cerr << "WARNING: very elongated particle, or wrong radius; some tests disabled\n";
            std::cerr << std::setprecision(16) << "  V  =" << V << "\n  R^3=" << R * R * R << "\n";
        }

        test_small_q(p, 1, 0, 0);
        test_small_q(p, -1, 0, 0);
        test_small_q(p, 0, 1, 0);
        test_small_q(p, 0, -1, 0);
        test_small_q(p, 0, 0, 1);
        test_small_q(p, 0, 0, -1);
        test_small_q(p, 1, 1, 0);
        test_small_q(p, 0, 1, 1);
        test_small_q(p, 1, 0, 1);
        test_small_q(p, 1, -1, 0);
        test_small_q(p, 0, 1, -1);
        test_small_q(p, -1, 0, 1);
        test_small_q(p, 1, 1, 1);
        test_small_q(p, .7, .8, .9);

        IFormFactor* clone = p->clone();
        EXPECT_EQ(clone->volume(), V);
        C3 q(.1, .2, complex_t(.3, .004));
        EXPECT_EQ(clone->formfactor_at_bottom(q), p->formfactor_at_bottom(q));
        delete clone;
    }
    double V, R;
};

TEST_F(FormFactorBasicTest, Name)
{
    Box ff(1., 2., 3.);
    EXPECT_EQ(ff.className(), "Box");
    EXPECT_EQ(ff.shapeName(), "Box");
}

TEST_F(FormFactorBasicTest, Pyramid2)
{
    EXPECT_THROW(Pyramid2(12, 14, -5, .8), std::runtime_error);
    EXPECT_THROW(Pyramid2(0, 1, 1, M_PI / 4), std::runtime_error);
    EXPECT_THROW(Pyramid2(1, 1, 1, 3 * M_PI / 4), std::runtime_error);

    double length = 12.;
    double width = 14.;
    double height = 5.;
    double alpha = 0.8;
    double tga = std::tan(alpha);
    double volume = height
                    * (length * width - (length + width) * height / tga
                       + 4.0 / 3.0 * height * height / (tga * tga));

    Pyramid2 particle(length, width, height, alpha);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(height, particle.topZ(&rot));

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, Bar)
{
    EXPECT_THROW(BarGauss(-100, -20, 4), std::runtime_error);
    EXPECT_THROW(BarGauss(100, 20, -4), std::runtime_error);
    // TODO EXPECT_THROW(BarGauss(0, 20, 4), std::runtime_error);
    // TODO EXPECT_THROW(BarGauss(100, 0, 4), std::runtime_error);
    // TODO EXPECT_THROW(BarGauss(100, 20, 0), std::runtime_error);
}

TEST_F(FormFactorBasicTest, Box)
{
    EXPECT_THROW(Box(-6, 5, -7), std::runtime_error);
    // TODO EXPECT_THROW(Box(6, 0, 7), std::runtime_error);

    double length = 6.;
    double height = 5.;
    double width = 7.;
    double volume = length * height * width;

    Box particle(length, width, height);
    EXPECT_EQ(3., particle.radialExtension());
    EXPECT_DOUBLE_EQ(volume, particle.volume());

    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(height, particle.topZ(&rot));

    EXPECT_EQ(0., particle.bottomZ(new RotationZ(17 * Units::deg)));
    EXPECT_EQ(height, particle.topZ(new RotationZ(39 * Units::deg)));

    EXPECT_NEAR(-width / 2, particle.bottomZ(new RotationX(90 * Units::deg)), 1e-12);
    EXPECT_NEAR(-length / 2, particle.bottomZ(new RotationY(90 * Units::deg)), 1e-12);

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, CantellatedCube)
{
    EXPECT_THROW(CantellatedCube(0, 0), std::runtime_error);
    EXPECT_THROW(CantellatedCube(10, -1), std::runtime_error);
    EXPECT_THROW(CantellatedCube(10, 6), std::runtime_error);

    double L = 10.;
    double t = 2.; // side length of removed trirectangular tetrahedron at each vertex
    double volume = L * L * L - 6 * L * t * t + 16 * t * t * t / 3;

    CantellatedCube particle(L, t);
    EXPECT_DOUBLE_EQ(particle.volume(), volume);
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(L, particle.topZ(&rot));

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, Cone)
{
    EXPECT_THROW(Cone(-6, -5, 0.8), std::runtime_error);
    // TODO EXPECT_THROW(Cone(6, 5, 0), std::runtime_error);
    EXPECT_THROW(Cone(6, 5, 2 * M_PI / 3), std::runtime_error);

    double radius = 6.;
    double height = 5.;
    double alpha = 0.8;
    double tga = std::tan(alpha);
    double HdivRtga = height / tga / radius;
    double volume = M_PI / 3. * tga * radius * radius * radius
                    * (1. - (1. - HdivRtga) * (1. - HdivRtga) * (1. - HdivRtga));

    Cone particle(radius, height, alpha);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(height, particle.topZ(&rot));

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, Pyramid6)
{
    EXPECT_THROW(Pyramid6(-6, -5, 0.8), std::runtime_error);
    // TODO EXPECT_THROW(Pyramid6(6, 5, 0), std::runtime_error);
    EXPECT_THROW(Pyramid6(6, 5, 2 * M_PI / 3), std::runtime_error);

    double base_edge = 6.;
    double height = 5.;
    double alpha = 0.8;
    double tga = std::tan(alpha);
    double HdivRtga = 2. * height / tga / base_edge / std::sqrt(3.);
    double volume = 3. / 4. * tga * base_edge * base_edge * base_edge
                    * (1. - (1. - HdivRtga) * (1. - HdivRtga) * (1. - HdivRtga));

    Pyramid6 particle(base_edge, height, alpha);
    EXPECT_DOUBLE_EQ(volume, particle.volume());

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, CosineRippleBox)
{
    EXPECT_THROW(CosineRippleBox(-20, 4, -100), std::runtime_error);
    // TODO EXPECT_THROW(CosineRippleBox(0, 4, 100), std::runtime_error);
    // TODO EXPECT_THROW(CosineRippleBox(20, 0, 100), std::runtime_error);
    // TODO EXPECT_THROW(CosineRippleBox(20, 4, 0), std::runtime_error);

    double width = 20.;
    double height = 4.;
    double length = 100.0;
    double volume = 0.5 * height * width * length;

    CosineRippleBox particle(length, width, height);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(height, particle.topZ(&rot));

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, Bipyramid4)
{
    double length = 10.;
    double height = 4;
    double height_ratio = .7;
    double alpha = 0.8;

    double tga = std::tan(alpha);
    double H2divLtga = height * 2. / length / tga;
    double ratioH2divLtga = height_ratio * height * 2. / length / tga;
    double volume = 1. / 6. * tga * length * length * length
                    * (2. - (1. - H2divLtga) * (1. - H2divLtga) * (1. - H2divLtga)
                       - (1. - ratioH2divLtga) * (1. - ratioH2divLtga) * (1. - ratioH2divLtga));

    Bipyramid4 particle(length, height, height_ratio, alpha);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(height * (1 + height_ratio), particle.topZ(&rot));

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, Cylinder)
{
    EXPECT_THROW(Cylinder(-3, -5), std::runtime_error);
    // TODO EXPECT_THROW(Cylinder(0, 5), std::runtime_error);

    double radius = 3.;
    double height = 5.;
    double volume = M_PI * radius * radius * height;

    Cylinder particle(radius, height);
    EXPECT_DOUBLE_EQ(volume, particle.volume());

    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(height, particle.topZ(&rot));

    RotationX rectX(90 * Units::deg);
    EXPECT_NEAR(-radius, particle.bottomZ(&rectX), 1e-13);
    EXPECT_NEAR(+radius, particle.topZ(&rectX), 1e-13);
    RotationY rectY(90 * Units::deg);
    EXPECT_NEAR(-radius, particle.bottomZ(&rectY), 1e-13);
    EXPECT_NEAR(+radius, particle.topZ(&rectY), 1e-13);

    RotationY invY(180 * Units::deg);
    EXPECT_NEAR(-height, particle.bottomZ(&invY), 1e-13);
    EXPECT_NEAR(0, particle.topZ(&invY), 1e-13);

    for (double gamma : {1.123, -2.34, 7.5, -9.})
        // 7.5deg is worst case for 24-vertex circle
        EXPECT_NEAR(-radius,
                    particle.bottomZ(new RotationEuler(0, 90 * Units::deg, gamma * Units::deg)),
                    3e-2); // TODO decrease epsilon after replacement of vertex-based approximation

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, HorizontalCylinder)
{
    EXPECT_THROW(HorizontalCylinder(-3, -5), std::runtime_error);

    double R = 3.; // radius
    double L = 5.; // length

    double b = -0;  // slice_bottom
    double t = 2.3; // slice_top

    double full_volume = M_PI * R * R * L;
    double trunc_volume =
        L
        * (t * sqrt(R * R - t * t) - b * sqrt(R * R - b * b) + R * R * (asin(t / R) - asin(b / R)));

    RotationZ rot(.42);

    // full

    HorizontalCylinder particle_full(R, L);
    EXPECT_DOUBLE_EQ(full_volume, particle_full.volume());

    EXPECT_EQ(0., particle_full.bottomZ(&rot));
    EXPECT_EQ(2 * R, particle_full.topZ(&rot));

    // truncated

    HorizontalCylinder particle_top(R, L, 0., R);
    EXPECT_NEAR(full_volume / 2, particle_top.volume(), 1e-7);

    HorizontalCylinder particle_bottom(R, L, -R, 0.);
    EXPECT_NEAR(full_volume / 2, particle_bottom.volume(), 1e-7);

    HorizontalCylinder particle_trunc(R, L, b, t);
    EXPECT_NEAR(trunc_volume, particle_trunc.volume(), 1e-7);

    EXPECT_EQ(0., particle_trunc.bottomZ(&rot));
    EXPECT_NEAR(t - b, particle_trunc.topZ(&rot), 1e-7);

    RotationX rot1(M_PI_2);
    EXPECT_NEAR(-R, particle_trunc.bottomZ(&rot1), 1e-7);
    EXPECT_NEAR(R, particle_trunc.topZ(&rot1), 1e-7);

    RotationX rot2(M_PI_4);
    EXPECT_NEAR(-R / sqrt(2.), particle_trunc.bottomZ(&rot2), 1e-7);
    EXPECT_NEAR(R, particle_trunc.topZ(&rot2), 1e-7);


    test_ff(&particle_trunc);
}

TEST_F(FormFactorBasicTest, Dodecahedron)
{
    EXPECT_THROW(Dodecahedron(-1), std::runtime_error);
    EXPECT_THROW(Dodecahedron(0), std::runtime_error);

    double edge = 3.;
    double volume = (15 + 7 * sqrt(5)) / 4 * pow(edge, 3);

    Dodecahedron particle(edge);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_NEAR(2 * 1.11352 * edge, particle.topZ(&rot), 1e-4);
    // height=2*inradius from web ressource

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, EllipsoidalCylinder)
{
    double radiusx = 3.;
    double radiusy = 5.;
    double height = 4;
    double volume = M_PI * radiusx * radiusy * height;

    EllipsoidalCylinder particle(radiusx, radiusy, height);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(height, particle.topZ(&rot));

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, FullSphere)
{
    EXPECT_THROW(FullSphere(-1), std::runtime_error);
    // TODO EXPECT_THROW(FullSphere(0), std::runtime_error);

    double radius = 5.;
    double volume = 4. / 3. * M_PI * radius * radius * radius;

    FullSphere particle(radius);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(2 * radius, particle.topZ(&rot));

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, FullSpheroid)
{
    EXPECT_THROW(FullSpheroid(-3, -5), std::runtime_error);
    // TODO EXPECT_THROW(FullSpheroid(0, 5), std::runtime_error);
    // TODO EXPECT_THROW(FullSpheroid(3, 0), std::runtime_error);
    double radius = 3.;
    double height = 5.;
    double volume = 2. / 3. * M_PI * radius * radius * height;

    FullSpheroid particle(radius, height);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(height, particle.topZ(&rot));

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, HemiEllipsoid)
{
    EXPECT_THROW(HemiEllipsoid(-6, -7, 5), std::runtime_error);

    double radiusx = 6.;
    double radiusy = 7.;
    double height = 5.;

    double volume = M_TWOPI * radiusx * radiusy * height / 3.;

    HemiEllipsoid particle(radiusx, radiusy, height);
    EXPECT_DOUBLE_EQ(volume, particle.volume());

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, Icosahedron)
{
    EXPECT_THROW(Icosahedron(-1), std::runtime_error);
    EXPECT_THROW(Icosahedron(0), std::runtime_error);

    double edge = 7.;
    double volume = 5 * (3 + sqrt(5)) / 12 * pow(edge, 3);

    Icosahedron particle(edge);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_NEAR(2 * 0.755761 * edge, particle.topZ(&rot), 1e-4);
    // height=2*inradius from web ressource

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, PlatonicOctahedron)
{
    EXPECT_THROW(PlatonicOctahedron(-1), std::runtime_error);
    EXPECT_THROW(PlatonicOctahedron(0), std::runtime_error);

    double edge = 7.;
    double volume = sqrt(2.) / 3 * pow(edge, 3);

    PlatonicOctahedron particle(edge);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_NEAR(edge * sqrt(2.), particle.topZ(&rot), 1e-4);

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, PlatonicTetrahedron)
{
    EXPECT_THROW(PlatonicTetrahedron(-1), std::runtime_error);
    EXPECT_THROW(PlatonicTetrahedron(0), std::runtime_error);

    double edge = 7.;
    double volume = sqrt(2.) / 12 * pow(edge, 3);

    PlatonicTetrahedron particle(edge);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_NEAR(edge * sqrt(2. / 3.), particle.topZ(&rot), 1e-4);

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, Prism3)
{
    EXPECT_THROW(Prism3(-4, -6), std::runtime_error);
    // TODO EXPECT_THROW(Prism3(0, 6), std::runtime_error);
    // TODO EXPECT_THROW(Prism3(4, 0), std::runtime_error);

    double height = 4.;
    double base_edge = 6.;
    double volume = sqrt(3.) / 4. * height * base_edge * base_edge;

    Prism3 particle(base_edge, height);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(height, particle.topZ(&rot));

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, Prism6)
{
    EXPECT_THROW(Prism6(-4, -3), std::runtime_error);
    // TODO EXPECT_THROW(Prism6(0, 3), std::runtime_error);
    // TODO EXPECT_THROW(Prism6(4, 0), std::runtime_error);

    double height = 4.;
    double base_edge = 3.;
    double volume = 3. * sqrt(3.) / 2. * height * base_edge * base_edge;

    Prism6 particle(base_edge, height);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(height, particle.topZ(&rot));

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, Pyramid4)
{
    EXPECT_THROW(Pyramid4(-10, -4, .8), std::runtime_error);
    EXPECT_THROW(Pyramid4(0, 4, .8), std::runtime_error);
    // TODO EXPECT_THROW(Pyramid4(10, 0, .8), std::runtime_error);
    EXPECT_THROW(Pyramid4(10, 4, 0), std::runtime_error);

    double base_edge = 10.;
    double height = 4.;
    double alpha = 0.8;
    double tga = std::tan(alpha);
    double H2divLtga = height * 2. / base_edge / tga;
    double volume = 1. / 6. * tga * base_edge * base_edge * base_edge
                    * (1. - (1. - H2divLtga) * (1. - H2divLtga) * (1. - H2divLtga));

    Pyramid4 particle(base_edge, height, alpha);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(height, particle.topZ(&rot));

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, SawtoothRippleBox)
{
    EXPECT_THROW(SawtoothRippleBox(-20, 4, -100, .3), std::runtime_error);
    // TODO EXPECT_THROW(SawtoothRippleBox(0, 4, 100, .3), std::runtime_error);
    // TODO EXPECT_THROW(SawtoothRippleBox(20, 0, 100, .3), std::runtime_error);
    // TODO EXPECT_THROW(SawtoothRippleBox(20, 4, 0, .3), std::runtime_error);
    // TODO EXPECT_THROW(SawtoothRippleBox(20, 4, 0, -.3), std::runtime_error);

    double width = 20.;
    double height = 4.;
    double length = 100.0;
    double d = 0.3; // asymmetry
    double volume = 0.5 * height * width * length;

    SawtoothRippleBox particle(length, width, height, d);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(height, particle.topZ(&rot));

    // test_ff( &particle ); WAITING: restore once radius returns the umkreis radius
}

TEST_F(FormFactorBasicTest, TruncatedCube)
{
    EXPECT_THROW(TruncatedCube(0, 0), std::runtime_error);
    EXPECT_THROW(TruncatedCube(10, -1), std::runtime_error);
    EXPECT_THROW(TruncatedCube(10, 6), std::runtime_error);

    double length = 15.;
    double t = 6.; // side length of removed trirectangular tetrahedron at each vertex
    double volume = length * length * length - 4. / 3. * t * t * t;

    TruncatedCube particle(length, t);
    EXPECT_DOUBLE_EQ(particle.volume(), volume);
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(length, particle.topZ(&rot));

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, TruncatedSphere)
{
    EXPECT_THROW(TruncatedSphere(-5, -3, 1.7), std::runtime_error);
    // TODO EXPECT_THROW(TruncatedSphere(5, 3, 0), std::runtime_error);
    EXPECT_THROW(TruncatedSphere(0, 3, 1.7), std::runtime_error);
    EXPECT_THROW(TruncatedSphere(5, 0, 1.7), std::runtime_error);

    double radius = 5.;
    double height = 3.;
    double HdivR = height / radius;
    double volume = M_PI / 3. * radius * radius * radius
                    * (3. * HdivR - 1. - (HdivR - 1.) * (HdivR - 1.) * (HdivR - 1.));

    TruncatedSphere particle(radius, height, 0);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(height, particle.topZ(&rot));

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, TruncatedSpheroid)
{
    EXPECT_THROW(TruncatedSpheroid(-5, -3, 1.7, 0), std::runtime_error);
    // TODO EXPECT_THROW(TruncatedSpheroid(5, 3, 0, 0), std::runtime_error);
    EXPECT_THROW(TruncatedSpheroid(0, 3, 1.7, 0), std::runtime_error);
    // TODO EXPECT_THROW(TruncatedSpheroid(5, 0, 1.7, 0), std::runtime_error);

    double radius = 3.;
    double height = 5.;
    double flattening = 1.5;
    double volume =
        M_PI * radius * height * height / flattening * (1. - height / (3. * flattening * radius));

    TruncatedSpheroid particle(radius, height, flattening, 0);
    EXPECT_DOUBLE_EQ(volume, particle.volume());
    RotationZ rot(.42);
    EXPECT_EQ(0., particle.bottomZ(&rot));
    EXPECT_EQ(height, particle.topZ(&rot));

    test_ff(&particle);
}

TEST_F(FormFactorBasicTest, Pyramid3)
{
    EXPECT_THROW(Pyramid3(-16, -4, .8), std::runtime_error);
    EXPECT_THROW(Pyramid3(0, 4, .8), std::runtime_error);
    // TODO EXPECT_THROW(Pyramid3(16, 0, .8), std::runtime_error);
    EXPECT_THROW(Pyramid3(16, 4, 0), std::runtime_error);

    double base_edge = 16.;
    double height = 4.;
    double alpha = 0.8;
    double tga = std::tan(alpha);
    double sqrt3H2divLtga = std::sqrt(3.) * 2. * height / base_edge / tga;
    double volume = tga / 24. * base_edge * base_edge * base_edge
                    * (1. - (1. - sqrt3H2divLtga) * (1. - sqrt3H2divLtga) * (1. - sqrt3H2divLtga));

    Pyramid3 particle(base_edge, height, alpha);
    EXPECT_DOUBLE_EQ(volume, particle.volume());

    test_ff(&particle);
}
